/*!
 * @file  Join_IAccessOperator
 * @brief implements Join_IAccessOperator
 *
 * @author GertG
 * @ingroup Join
 *
 * @par last changed by:
 * <br>
 * $Author: d026026 $ $DateTime: 2004/09/10 15:00:56 $
 *
 * @sa Join_IAccessOperator.hpp
 */
/*

    ========== licence begin  GPL
    Copyright (c) 2002-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end


*/

#include "SAPDB/Join/Join_IAccessOperator.hpp"

#include "SAPDB/Trace/Trace_Entry.hpp"
#include "Trace/Trace_MessBlockEntry.hpp"
#include "SAPDBCommon/SAPDB_Types.hpp"
#include "ggg00.h"
#include "hgg01_3.h"
#include "ggg11.h"
#include "hgg04.h"
#include "gsp00.h"
#include "hkb71.h"
#include "hta01.h"


/* ******************** PUBLIC MEMBERS ********************* */
/*!
 * @param acv [in] global context
 */
Join_IAccessOperator::Join_IAccessOperator(
    tak_all_command_glob& acv,
    const SAPDB_UInt2 tabno)
    : IOperator(acv),
      m_AccessDesc(acv), m_NullRec(0), m_TableNo(tabno),
      m_maxRowRequest(SAPDB_MAX_INT4)
{
    SAPDBTRACE_METHOD_DEBUG( "Join_IAccessOperator::ctor", Join_Trace, 1 );

    SAPDBTRACE_WRITELN(
        Join_Trace, 5,
        "\nstrat reclen: " << GetRecordLength() );

    if ( g01vtrace.vtrAll_gg00 )
    {
        // tracing
        tgg11_KbTabKeyTrace _mess_trace;
        _mess_trace.kbtkTabHead_gg11.kbtHead_gg11.kbMessType_gg11  =
            m_AccessDesc.MBlock().mb_type();
        _mess_trace.kbtkTabHead_gg11.kbtHead_gg11.kbMessType2_gg11 =
            m_AccessDesc.MBlock().mb_type2();
        _mess_trace.kbtkTabHead_gg11.kbtHead_gg11.kbSubtrans_gg11 =
            m_AccessDesc.MBlock().mb_trns()->trSubtransId_gg00;
        _mess_trace.kbtkTabHead_gg11.kbtTabId_gg11 =
            m_AccessDesc.MBlock().mb_qual()->mtree().fileTabId_gg00();
        _mess_trace.kbtkTabHead_gg11.kbtKeyLen_gg11 = 0;
        SAPDB_UInt4 __len = sizeof( _mess_trace.kbtkTabHead_gg11 );
        if ( m_AccessDesc.MBlock().mb_trns()->trError_gg00 != e_move_error )
        {
            Trace_CommonEntry Entry(
                *m_AccessDesc.MBlock().mb_trns(),
                tgg00_Debug::fromConst(kb),
                tgg00_VtraceType::fromConst(vttKbFunctions_egg00),
                (tsp00_BytePtr) &_mess_trace, __len );
        }
    }

    if ( Join_Trace.TracesLevel( 1 ) )
    {
        tgg00_Debug Layer = tgg00_Debug::fromConst( ak_messbuf );

        Trace_MessBlockEntry Entry(
            m_acv.a_transinf.tri_trans, m_acv.a_mblock, Layer,
            tgg00_VtraceType::fromConst(ak_send) );
    }

    invalidate_keys();
}

/*************************************************************************/

void Join_IAccessOperator::SetMaximumRowsReadAtATime(
    const SAPDB_Int4 maxRowRequest)
{
    m_maxRowRequest = maxRowRequest;
}

/*************************************************************************/

SAPDB_Int4 Join_IAccessOperator::GetMaximumRowsReadAtATime() const
{
    return m_maxRowRequest;
}

/*************************************************************************/

tgg00_BasisError Join_IAccessOperator::GetNullRecord( tgg00_Rec*& recptr )
{
    SAPDBTRACE_METHOD_DEBUG( "Join_IAccessOperator::GetNullRecord", Join_Trace, 1 );

    if ( !m_NullRec && m_AccessDesc.HasOutputColumns() )
    {
        SAPDBTRACE_WRITELN( Join_Trace, 7, "create NULL record" );
        
        if (!( m_NullRec = (tgg00_Rec*) 
                    this->GetAlloc().Allocate( this->GetRecordLength() )))
            return e_no_more_memory;

        SAPDBTRACE_IF( Join_Trace, 3, SAPDB_MemFillNoCheck(m_NullRec, ':', this->GetRecordLength()) );
        tgg00_StackDesc _output_desc = m_AccessDesc.MBlock().mb_qual()->mstack_desc();
        tgg00_StackList *_output = 0; 
        
        if ( _output = (tgg00_StackList*) 
                    this->GetAlloc().Allocate( ((*m_AccessDesc.MBlock().mb_st())[m_AccessDesc.MBlock().mb_qual()->mqual_pos()-1].epos() - 2) * sizeof( tgg00_StackEntry ) ))
        {
            SAPDBTRACE_WRITELN( Join_Trace, 0, "allocate _output @ " << (void*)_output );
            SAPDB_MemCopyNoCheck( _output, (tgg00_StackEntry *)m_AccessDesc.MBlock().mb_st() + ( m_AccessDesc.MBlock().mb_qual()->mqual_pos() - 1 + 1 ), 
                     ((*m_AccessDesc.MBlock().mb_st())[m_AccessDesc.MBlock().mb_qual()->mqual_pos()-1].epos() - 2) * sizeof( tgg00_StackEntry ) );
            _output_desc.mst_addr()         = _output;
            _output_desc.mqual_cnt          = 
                    ((*m_AccessDesc.MBlock().mb_st())[m_AccessDesc.MBlock().mb_qual()->mqual_pos()-1].epos() - 2);
            _output_desc.mqual_pos = 1;
            _output_desc.mst_max()          = _output_desc.mqual_cnt;
            _output_desc.mfirst_free()      = _output_desc.mqual_cnt + 1;
            _output_desc.mst_optimize_pos() = 0;
            _output_desc.mstrat_pos         =  _output_desc.mstrat_cnt     = 0;
            _output_desc.mresqual_pos()     =  _output_desc.mresqual_cnt() = 0;
            _output_desc.mcol_pos           =  _output_desc.mcol_cnt       = 0;
            _output_desc.mmult_pos          =  _output_desc.mmult_cnt      = 0;
            _output_desc.mview_pos          =  _output_desc.mview_cnt      = 0;
            _output_desc.mupd_pos           =  _output_desc.mupd_cnt       = 0;
            _output_desc.mlink_pos          =  _output_desc.mlink_cnt      = 0;
            _output_desc.mstring_pos        =  _output_desc.mstring_cnt    = 0;
            _output_desc.mtrigger_pos       =  _output_desc.mtrigger_cnt   = 0;
            _output_desc.minvqual_pos       =  _output_desc.minvqual_cnt   = 0;
            
            // adjust column stack entries 
            for ( int _i = 0; _i < _output_desc.mqual_cnt; ++_i )
            {
                switch ( (*_output)[ _i ].etype() )
                {
                    case st_fixkey: 
                    case st_fixcol: 
                    case st_varkey: 
                    case st_varcol: 
                    case st_fixinv: 
                    case st_varinv: 
                    case st_varlongchar: 
                    case st_column: 
                        (*_output)[ _i ].etype().becomes( st_value );
                        // NULL value asured through a689search_sequence()
                        (*_output)[ _i ].epos()  = m_AccessDesc.MBlock().mb_data_len();
                        (*_output)[ _i ].elen_var() = 1;
                        (*_output)[ _i ].ecol_tab()[0] = 0;
                        (*_output)[ _i ].ecol_tab()[1] = 0;
                        break;
                }
            }
            
            SAPDBTRACE_IF( Join_Trace, 3, 
            t01stackdesc( td_always, "copied outpu", _output_desc.mst_addr(), _output_desc ));
            
            tgg00_StackEntry *_err_st_entry;
            pasbool _unqualified = false;

            tgg00_SelectFieldsParam _selfields;
            g04init_select_fields( _selfields, 
                    reinterpret_cast<tsp00_MoveObjPtr>(m_AccessDesc.MBlock().mb_data()),
                    m_AccessDesc.MBlock().mb_data_size(), 
                    m_AccessDesc.MBlock().mb_valuearr(),
                    m_AccessDesc.MBlock().mb_validx_max(),
                    m_acv.a_work_st_addr, 
                    m_acv.a_work_st_max, 
                    m_acv.a_work_buf_addr, 
                    m_acv.a_work_buf_size,
                    m_acv.a_sqlmode );
            _selfields.sfp_bd_mess_type().becomes( m_select );
            _selfields.sfp_result_wanted() = true;
            _selfields.sfp_first_qual()    = true;
            _selfields.sfp_rows_read()     = 0;
            _selfields.sfp_rows_qual()     = 0;
            _selfields.sfp_acv_addr()      = reinterpret_cast<tsp00_Addr>( &m_acv );
            _selfields.sfp_oldrec_addr()   = 0;
            _selfields.sfp_primkey_addr()  = 0;
            _selfields.sfp_result_length() = 0; 
            _selfields.sfp_m_result_len()  = 0;
            _selfields.sfp_m_result_cnt()  = 1;
            _selfields.sfp_m_result_size() = this->GetRecordLength();
            _selfields.sfp_m_result_addr() = reinterpret_cast<tsp00_MoveObjPtr>(m_NullRec);
            _selfields.sfp_rec_addr()    = reinterpret_cast<tsp00_BufAddr>(m_NullRec);
            _selfields.sfp_rec_len()     = m_NullRec->recLen_gg00()    = this->GetRecordLength(); 
            _selfields.sfp_rec_key_len() = m_NullRec->recKeyLen_gg00() = this->GetKeyLength();
            m_AccessDesc.MBlock().mb_trns()->trError_gg00 = e_ok;
            
            k71qual_handling( *m_AccessDesc.MBlock().mb_trns(), _selfields, 
                    false, false, _output_desc,
                    _err_st_entry, _unqualified );
            SAPDBTRACE_IF( Join_Trace, 3, t01buf( td_always, m_NullRec, 1, this->GetRecordLength() ) );
            
            this->GetAlloc().Deallocate( _output );
        }
        else
            return e_no_more_memory;
        
    }
    recptr = m_NullRec;
    return e_ok;
}

