/*!
 * @file
 * @brief implements Join_JoinOperator2
 *
 * @author GertG
 * @ingroup Join
 *
 * @par last changed by:
 * <br>
 * $Author: d026026 $ $DateTime: 2004/09/10 15:00:56 $
 *
 * @sa Join_JoinOperator2.hpp
 */
/*

    ========== licence begin  GPL
    Copyright (c) 2002-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end


*/
#include "gsp00.h"
#include "ggg00.h"
#include "ggg07.h"
#include "gkb07.h"
#include "SAPDBCommon/Tracing/SAPDBTrace_Hex.hpp"
#include "SAPDBCommon/Tracing/SAPDBTrace_HexBuf.hpp"
#include "SAPDBCommon/ErrorsAndMessages/SAPDBErr_Assertions.hpp"
#include "SAPDBCommon/SAPDB_RangeCode.hpp"
#include "SAPDBCommon/SAPDB_MemCopyMove.hpp"
#include "Join/Join_JoinOperator2.hpp"
#include "Join/Join_TableAccessOperator.hpp"
#include "Join/Join_IndexAccessOperator.hpp"
#include "Join/Join_HashAccessAllocator.hpp"
#include "Join/Join_HashAccessOperator.hpp"
#include "hak682.h"
#include "hak683.h"
#include "hak01.h"
#include "hak101.h"
#include "hak07.h"
#include "hak70.h"
#include "hkb74.h"
#include "hsp30.h"
#include "hgg02.h"
#include "hgg20.h"
#include "hta01.h"
#include "hta01_3.h"
#include "hsp49.h"
#include "hkb71.h"
#include "hgg04.h"
#include "hbd01.h"
#include "hbd01_1.h"
#include "hbd07.h"
#include "hkb720.h"
#include "hbd21.h"

extern "C"
void vdebug_break( SAPDB_Int4 );

/* ******************** PUBLIC MEMBERS ********************* */
/*! 
 * @param acv [in] global context
 * @param dmli [in] statement context
 * @param parsk [in] parskey to be used
 * @param jvrec [in] join view record
 * @param use_old_rescnt [in] delete 'CreateFile' flag of result file?
 * @param del_parsinfos [in] delete parse infos?
 * @param sequence [in] pointer to join table sequence array
 * @param righttabidx [in] table index of right table (start with idx 1)
 */
Join_JoinOperator2::Join_JoinOperator2(
        tak_all_command_glob        &acv,
        tak_dml_info                &dmli,
        const tak_parskey           &parsk,
        const tak68_joinview_rec    &jvrec,
        const pasbool               use_old_rescnt,
        const pasbool               del_parsinfos,
        const tak68_sequence        &sequence,
        const SAPDB_Int4            righttabidx ):
    IOperator(acv), 
    m_LeftOp(0), m_RightOp(0), 
    m_JoinRec(0), m_NullRec(0), m_RightAuxRec(0),
    m_Records(0), 
    m_Filter(0), 
    m_OutputInfo(0), 
    m_JoinRecBufferSize(0), 
    m_AuxRightOuterKeyLen(0),
    m_RightRecConvArrCnt(0), 
    m_RightAccessInvLen(sequence[righttabidx-1].jos_invlen),
    m_LastRightDefBytePos(0), 
    m_OutputInfoSize(0),
    m_RightTableIdx(righttabidx),
    m_JoinSequence(sequence),
    // m_RightRecConvArr
    // m_JoinInfo
    // m_FilterDesc
    // m_SelFields
    m_AuxRightOuterJoinFile(b01niltree_id), 
    // m_AuxRightOuterJoinFilePos, 
    // m_AuxRightOuterJoinKey
    m_d_joins(dmli.d_joins), 
    m_RightAccessStrat(sequence[righttabidx-1].jos_joinstrat),
    m_TableForRightOuterJoinScanStat(never_scanned), 
    m_IsTopJoinOp(righttabidx==dmli.d_cntfromtab), 
    m_GetLeftTupel(true), 
    m_LeftRecChanged(false), 
    m_LeftRecJoined(false),
    m_IsDescendingIndex(false),
    m_AppendRightOuterJoinRecords(false),
    m_OutputFilter(false)
{
    // precondition: acv.a_mblock contains messblock for right table access op.
    // guarantee destructor
    m_SelFields.sfp_data_addr() = 0;
    SAPDBERR_ASSERT_STATE( m_RightTableIdx > 1 );

    if ( 0 == m_acv.a_returncode )
    {
        SAPDBTRACE_METHOD_DEBUG( "Join_JoinOperator2::ctor", Join_Trace, 1 );
        
        SAPDBTRACE_WRITELN( Join_Trace, 3, "right table idx : " << m_RightTableIdx 
                << "\taccess : " << sequence[m_RightTableIdx-1].jos_joinstrat );
        
        tgg00_BasisError _b_err;

        _b_err = operator_init();

        if ( e_ok == _b_err )
        {
            SAPDBTRACE_WRITELN( Join_Trace, 3, "create right AccessOp" );
            // create new right AccessOperator with actual acv.a_mblock
            m_RightOp = this->AccessOperatorFactory(
                acv,
                sequence[m_RightTableIdx-1].jos_table_buffer,
                sequence[m_RightTableIdx-1].jos_expected_table_rec_reads,
                sequence[m_RightTableIdx-1].jos_source);
            if ( ! m_RightOp )
                _b_err = e_no_more_memory;
            
            if ( 0 == acv.a_returncode && e_ok == _b_err )
            {
                // get next acv.a_mblock
                a682j_one_record( acv, dmli, m_RightTableIdx-1, 
                        sequence[m_RightTableIdx-1-1].jos_source, parsk, 
                        jvrec, use_old_rescnt, del_parsinfos );
                
                if ( 0 == acv.a_returncode )
                {
                    if ( 2 == m_RightTableIdx )
                    {
                        SAPDBTRACE_WRITELN( Join_Trace, 3, "create left AccessOp" );
                        // last join operator, therefore create left AccessOperator
                        m_LeftOp = this->AccessOperatorFactory(
                            acv,
                            sequence[m_RightTableIdx-1-1].jos_table_buffer,
                            sequence[m_RightTableIdx-1-1].jos_expected_table_rec_reads,
                            sequence[m_RightTableIdx-1-1].jos_source);
                        if ( ! m_LeftOp )
                            _b_err = e_no_more_memory;
                    }
                    else
                    {
                        SAPDBTRACE_WRITELN( Join_Trace, 3, "create left JoinOp" );
                        // create new left JoinOperator
                        if (!(m_LeftOp = new(this->GetAlloc()) 
                                    Join_JoinOperator2( acv, dmli, 
                                        parsk, jvrec, use_old_rescnt, 
                                        del_parsinfos, 
                                        sequence, m_RightTableIdx-1 )))
                            _b_err = e_no_more_memory;
                    }
                }
            }
        }

        if ( m_JoinInfo.jn_rightoj )
        {
            SAPDBERR_ASSERT_STATE( !a70glob_join_strats.includes( m_RightAccessStrat ) );

            if (!( m_RightAuxRec = (tgg00_Rec*) 
                        this->GetAlloc().Allocate( m_RightOp->GetRecordLength() )))
                _b_err = e_no_more_memory;
        }

        // clean up
        if ( _b_err != e_ok )
            a07_b_put_error( acv, _b_err, 1 );
    }
}

/*!
 *
 */
Join_JoinOperator2::~Join_JoinOperator2()
{
    SAPDBTRACE_METHOD_DEBUG( "Join_JoinOperator2::dtor", Join_Trace, 1 );

    // release space for data part
    SAPDBTRACE_WRITELN( Join_Trace, 8, "deallocate m_SelFields.sfp_data_addr @ " << (void*)m_SelFields.sfp_data_addr() );
    if ( m_SelFields.sfp_data_addr() ) this->GetAlloc().Deallocate( m_SelFields.sfp_data_addr() );
    // release space for stack entries
    SAPDBTRACE_WRITELN( Join_Trace, 8, "deallocate m_Filter @ " << (void*)m_JoinRec );
    if ( m_Filter  ) this->GetAlloc().Deallocate( m_Filter );
    SAPDBTRACE_WRITELN( Join_Trace, 8, "deallocate m_LeftOp @ " << (void*)m_LeftOp );
    if ( m_LeftOp  ) destroy( m_LeftOp,  this->GetAlloc() );
    SAPDBTRACE_WRITELN( Join_Trace, 8, "deallocate m_RightOp @ " << (void*)m_RightOp );
    if ( m_RightOp ) destroy( m_RightOp, this->GetAlloc() );
    SAPDBTRACE_WRITELN( Join_Trace, 8, "deallocate m_JoinRec @ " << (void*)m_JoinRec );
    if ( m_JoinRec ) this->GetAlloc().Deallocate( m_JoinRec );
    SAPDBTRACE_WRITELN( Join_Trace, 8, "deallocate m_NullRec @ " << (void*)m_NullRec );
    if ( m_NullRec ) this->GetAlloc().Deallocate( m_NullRec );
    SAPDBTRACE_WRITELN( Join_Trace, 8, "deallocate m_RightAuxRec @ " << (void*)m_RightAuxRec );
    if ( m_RightAuxRec ) this->GetAlloc().Deallocate( m_RightAuxRec );
    SAPDBTRACE_WRITELN( Join_Trace, 8, "deallocate m_OutputInfo @ " << (void*)m_OutputInfo );
    if ( m_OutputInfo ) this->GetAlloc().Deallocate( m_OutputInfo );
    SAPDBTRACE_WRITELN( Join_Trace, 8, "deallocate m_Records @ " << (void*)m_Records );
    if ( m_Records ) this->GetAlloc().Deallocate( m_Records );
    // make save forgotten Close()
    if ( m_JoinInfo.jn_rightoj && NIL_PAGE_NO_GG00 != m_AuxRightOuterJoinFile.fileRoot_gg00() )
        b01destroy_file( *m_acv.a_mblock.mb_trns(), m_AuxRightOuterJoinFile );
}

/*!
 *
 */
void Join_JoinOperator2::Open()
{
    SAPDBTRACE_METHOD_DEBUG( "Join_JoinOperator2::Open", Join_Trace, 1 );
    // open outer stream
    m_LeftOp->Open();
    init();
    IOperator::Open();
}

/*! 
 * @param startkeys [in] restrict record stream to records with key greater/equal than startkey
 * @param stopkeys [in] restrict record stream to records with key lower/equal than stopkey
 */
void Join_JoinOperator2::Open(
        const Join_TwoKeys &startkeys, 
        const Join_TwoKeys &stopkeys )
{
    SAPDBTRACE_METHOD_DEBUG( "Join_JoinOperator2::Open", Join_Trace, 1 );
    // open outer stream
    m_LeftOp->Open( startkeys, stopkeys );
    init();
    IOperator::Open( startkeys, stopkeys );
}


/*! 
 * @param recptr [in/out] pointer to memory should be filled with record
 * @return information code <tt>[e_ok, e_no_next_record]</tt> / error code
 */
tgg00_BasisError Join_JoinOperator2::Next( tgg00_Rec* &recptr )
{
    SAPDBTRACE_METHOD_DEBUG( "Join_JoinOperator2::Next", Join_Trace, 1 );
    tgg00_BasisError _b_err;
    SAPDB_Bool _qualified;

    do 
    {
        if ( e_ok == ( _b_err = this->next( m_Records ) ))
            create_output( m_JoinRec, m_Records );
        if ( m_Filter && e_ok == _b_err )
            _qualified = filter( m_JoinRec, _b_err );
        else
            _qualified = true;
    } 
    while ( e_ok == _b_err && !_qualified );
    
    recptr = m_JoinRec;
    
    return _b_err;
}

/*!
 *
 */
void Join_JoinOperator2::Close()
{
    SAPDBTRACE_METHOD_DEBUG( "Join_JoinOperator2::Close", Join_Trace, 1 );
    // close outer stream
    if ( m_LeftOp->IsOpened() )
        m_LeftOp->Close();

    // close inner stream
    if ( m_RightOp->IsOpened() )
        m_RightOp->Close();
    
    if ( IsOpened() && m_JoinInfo.jn_rightoj && 
         NIL_PAGE_NO_GG00 != m_AuxRightOuterJoinFile.fileRoot_gg00() )
    {
        b01destroy_file( *m_acv.a_mblock.mb_trns(), m_AuxRightOuterJoinFile );
        m_AuxRightOuterJoinFile.fileRoot_gg00() = NIL_PAGE_NO_GG00;
    }
    IOperator::Close();
}

/*!
 * @param record [in/out] record to be filled
 */
tgg00_BasisError Join_JoinOperator2::GetNullRecord( tgg00_Rec* &recptr ) 
{
    SAPDBTRACE_METHOD_DEBUG( "Join_JoinOperator2::GetNullRecord", Join_Trace, 1 );
    tgg00_BasisError _b_err = e_ok;

    if ( ! m_NullRec )
    {
        SAPDBTRACE_WRITELN( Join_Trace, 7, "create NULL record" );

        // reserve extra space (failure tolerance)
        if (!( m_NullRec = (tgg00_Rec*) 
                    this->GetAlloc().Allocate( m_JoinRecBufferSize )))
            return e_no_more_memory;
        SAPDBTRACE_WRITELN( Join_Trace, 3, "allocate m_NullRec @ " << 
                (void*)m_NullRec << " with length of " << m_JoinRecBufferSize );
        SAPDBTRACE_IF( Join_Trace, 3, 
                SAPDB_MemFillNoCheck( m_NullRec, ':', m_JoinRecBufferSize ) );

        tgg00_Rec** _tmprecs;
        if ( ! ( _tmprecs = 
               (tgg00_Rec**) this->GetAlloc().
               Allocate( m_RightTableIdx * sizeof(tgg00_Rec*))) )
            return e_no_more_memory;

        if ( e_ok != ( _b_err = m_LeftOp->get_null_record( _tmprecs ) ))
            return _b_err;
        if ( e_ok != ( _b_err = m_RightOp->get_null_record( _tmprecs ) )) 
            return _b_err;
        create_output( m_NullRec, _tmprecs );

        if ( m_Filter && m_OutputFilter ) 
            (void) filter( m_NullRec, _b_err );
        
        this->GetAlloc().Deallocate( _tmprecs );
    }
    recptr = m_NullRec;
    SAPDBTRACE_IF( Join_Trace, 5, 
        t01buf( td_always, recptr, 1, recptr->recLen_gg00() <= 300 ? 
            recptr->recLen_gg00() : 300 ));
    return _b_err;
}

/* ******************* PROTECTED MEMBERS ******************* */

/*! 
 * @param recptr [in/out] pointer to memory should be filled with record
 * @return information code <tt>[e_ok, e_no_next_record]</tt> / error code
 */
tgg00_BasisError Join_JoinOperator2::next( tgg00_Rec* *records )
{
    SAPDBTRACE_METHOD_DEBUG( "Join_JoinOperator2::next(<vector>)", Join_Trace, 1 );
    tgg00_BasisError _b_err = e_ok;

    do
    {
        if ( ! m_AppendRightOuterJoinRecords )
        {
            if ( m_GetLeftTupel )
            {
                m_GetLeftTupel    = false;
                m_LeftRecChanged  = false;
                m_LeftRecJoined = false;
                SAPDBTRACE_WRITELN( Join_Trace, 3, "<<get next left record (" << m_RightTableIdx  << ")" );
                _b_err = m_LeftOp->next( records );
                if ( e_no_next_record == _b_err )
                {
                    if ( m_JoinInfo.jn_rightoj )
                    {
                        m_AppendRightOuterJoinRecords = true;
                    }
                    else
                    {
                        // don't close left file to ensure repeatable Next() operation
                        //m_LeftOp->Close();

                        // inner join and no left record 
                        SAPDBTRACE_WRITELN( Join_Trace, 5, "end of Next @ " << (void *)this );
                        SAPDBTRACE_WRITELN( Join_Trace, 5, "because of left e_no_next_record" );
                        return e_no_next_record;
                    }
                }
                else if ( e_ok == _b_err )
                {
                    SAPDBTRACE_WRITELN( Join_Trace, 3, "left record (len=" << m_LeftOp->GetRecordLength() << ") @ " << (void *) this << " is: " );
                    if ( a70glob_join_strats.includes( m_RightAccessStrat ) )
                        // set m_Startkeys, m_Stopkeys
                        _b_err = prepare_right_key( records );
                }
            }

            if ( e_ok == _b_err )
            {
                SAPDB_Bool _joining;
                
                // open right operator
                if ( ! m_RightOp->IsOpened() )
                {
                    if ( a70glob_join_strats.includes( m_RightAccessStrat ) )
                    {
                        if ( m_JoinInfo.jn_rightoj && never_scanned == m_TableForRightOuterJoinScanStat )
                        {
                            SAPDBTRACE_WRITELN( Join_Trace, 5, "scan for RIGHT OUTER JOIN" );
                            m_RightOp->Open();
                        }
                        else
                            m_RightOp->Open( m_Startkeys, m_Stopkeys );
                    }
                    else
                        // use one table strategy
                        m_RightOp->Open();
                }

                do 
                {
                    SAPDBTRACE_WRITELN( Join_Trace, 3, "<<get next right record from " << m_RightTableIdx << ". tab (" << m_RightAccessStrat << ")");
                    _b_err = m_RightOp->next( records );
                    if ( e_ok == _b_err )
                    {
                        if ( never_scanned == m_TableForRightOuterJoinScanStat )
                            m_TableForRightOuterJoinScanStat = while_scanning;

                        if ( _joining = join( records, _b_err ) ) 
                            m_LeftRecJoined = true;
                            
                        if ( m_JoinInfo.jn_rightoj )
                        {
                            if ( !_joining && while_scanning == m_TableForRightOuterJoinScanStat )
                            {
                                /* !!! RIGHT OUTER JOIN !!! */
                                // add record with no join to auxiliary file
                                SAPDBTRACE_WRITELN( Join_Trace, 3, "add to ROJ aux file" << " m_AuxRightOuterKeyLen: " << m_AuxRightOuterKeyLen );

                                (*(records+(m_RightTableIdx-1)))->recLen_gg00()    = m_RightOp->GetRecordLength();
                                (*(records+(m_RightTableIdx-1)))->recKeyLen_gg00() = m_AuxRightOuterKeyLen;
                                (*(records+(m_RightTableIdx-1)))->recVarcolOffset_gg00() = 0;
                                (*(records+(m_RightTableIdx-1)))->recVarcolCnt_gg00()    = 0;
                                
                                SAPDBTRACE_IF( Join_Trace, 4, 
                                t01buf( td_always, (*(records+(m_RightTableIdx-1))), 1, (*(records+(m_RightTableIdx-1)))->recLen_gg00() <= 300 ? (*(records+(m_RightTableIdx-1)))->recLen_gg00() : 300 ));
                                b07cadd_record( m_acv.a_transinf.tri_trans, m_AuxRightOuterJoinFile, *(*(records+(m_RightTableIdx-1))) );
                                _b_err = m_acv.a_mblock.mb_trns()->trError_gg00;
                            }

                            if ( _joining && already_scanned == m_TableForRightOuterJoinScanStat )
                            {
                                SAPDBTRACE_WRITELN( Join_Trace, 3, "delete from ROJ aux file" );
                                SAPDBTRACE_IF( Join_Trace, 4, 
                                t01buf( td_always, &((*(records+(m_RightTableIdx-1)))->recBuf_gg00()), 1, m_RightOp->GetRecordLength() <= 300 ? m_RightOp->GetRecordLength() : 300 ));
                                
                                tgg00_Lkey _key;
                                _key.keyLen_gg00() = m_AuxRightOuterKeyLen;
                                SAPDB_MemCopyNoCheck( &_key.keyVal_gg00(), 
                                        &((*(records+(m_RightTableIdx-1)))->recKey_gg00().keyVal_gg00()),
                                        m_AuxRightOuterKeyLen );
                                SAPDBTRACE_IF( Join_Trace, 3, 
                                    t01key( td_always, "key         ", _key ));
                                b07cdel_record( m_acv.a_transinf.tri_trans, m_AuxRightOuterJoinFile, _key );
                                // ignore already joined records
                                if ( e_key_not_found == m_acv.a_mblock.mb_trns()->trError_gg00 )
                                    m_acv.a_mblock.mb_trns()->trError_gg00 = e_ok;
                                _b_err = m_acv.a_mblock.mb_trns()->trError_gg00;
                            }
                        }
                    }

                    // test for command canceling
                    if ( m_acv.a_transinf.tri_trans.trRteCommPtr_gg00->to_cancel )
                        _b_err = e_cancelled;
                }
                while ( e_ok == _b_err && !_joining );

                if ( e_no_next_record == _b_err )
                {
                    // no more right records -> switch inner state
                    m_GetLeftTupel = true;
                    if ( while_scanning == m_TableForRightOuterJoinScanStat )
                        m_TableForRightOuterJoinScanStat = already_scanned;
                    if ( !m_LeftRecJoined && m_JoinInfo.jn_leftoj )
                    {
                        SAPDBTRACE_WRITELN( Join_Trace, 5, "append LEFT OUTER record " );
                        /* !!! LEFT OUTER JOIN !!! */
                        _b_err = m_RightOp->get_null_record( records );
                    }
                    m_RightOp->Close();
                }
            }
        }

        if ( m_AppendRightOuterJoinRecords )
        {
            SAPDBTRACE_WRITELN( Join_Trace, 5, "append RIGHT OUTER record " );
            
            // tie NULL record on left side
            _b_err = m_LeftOp->get_null_record( records );
            if  ( e_ok != _b_err ) return _b_err;

            if ( never_scanned == m_TableForRightOuterJoinScanStat )
            {
                SAPDBTRACE_WRITELN( Join_Trace, 5, "right tab never scanned" );
                /* !!! RIGHT OUTER JOIN for empty left table !!! */
                // left table delivers no records, therfore m_AuxRightOuterJoinFile
                // is empty but we have to select whole right table
                
                if ( ! m_RightOp->IsOpened() ) m_RightOp->Open();

                _b_err = m_RightOp->next( records );

                if  ( e_no_next_record == _b_err)
                {
                    m_RightOp->Close();
                    m_GetLeftTupel = true;
                }
            }
            else
            {
                tgg00_BdSetResultRecord _set_result;
                _set_result.bd_key_check_len = 0;
                _set_result.bd_max_rec_cnt   = 1;
                _set_result.bd_max_fill_len  = m_RightOp->GetRecordLength();
                _set_result.bd_next          = true;
                _set_result.bd_drop_page     = false;
                
                // fill m_RightAuxRec with unmatched right record 
                b07cnext_record( m_acv.a_transinf.tri_trans, 
                        m_AuxRightOuterJoinFile, 
                        m_AuxRightOuterJoinKey, _set_result, 
                        m_AuxRightOuterJoinFilePos, 
                        m_RightAuxRec );
                if ( e_key_not_found == m_acv.a_transinf.tri_trans.trError_gg00 )
                    m_acv.a_transinf.tri_trans.trError_gg00 = e_ok;
                if ( e_ok == m_acv.a_transinf.tri_trans.trError_gg00 )
                {
                    SAPDBTRACE_IF( Join_Trace, 3, 
                    t01buf( td_always, m_RightAuxRec, 1, 
                        m_RightAuxRec->recLen_gg00() <= 300 ? 
                        m_RightAuxRec->recLen_gg00() : 300 ));

                    *(records+(m_RightTableIdx-1)) = m_RightAuxRec;
                }
                _b_err = m_acv.a_mblock.mb_trns()->trError_gg00;
            }

            if ( e_no_next_record == _b_err )
            {
                // don't close left file to ensure repeatable Next() operation
                // m_LeftOp->Close();

                // RIGHT OUTER JOIN and no more right records
                SAPDBTRACE_WRITELN( Join_Trace, 5, 
                        "end of Next @ " << (void *)this );
                SAPDBTRACE_WRITELN( Join_Trace, 5, 
                        "because of ROJ e_no_next_record" );
                return e_no_next_record;
            }
        } // m_AppendRightOuterJoinRecords
    }
    while ( e_no_next_record == _b_err );

    // reach this if _b_err == e_ok or sudden error situation
    
    SAPDBTRACE_WRITELN( Join_Trace, 5, "end of Next @ " << (void *)this );
    SAPDBTRACE_WRITELN( Join_Trace, 5, "return " << 
            SAPDBTrace::BasisError(_b_err) );
    
    return _b_err;
}

/*!
 * @param record [in/out] record to be filled
 */
tgg00_BasisError Join_JoinOperator2::get_null_record( tgg00_Rec** records ) 
{
    SAPDBTRACE_METHOD_DEBUG( "Join_JoinOperator2::get_null_record(<vector>)", Join_Trace, 1 );
    tgg00_BasisError _b_err;

    if ( e_ok != ( _b_err = m_LeftOp->get_null_record( records )) )
        return _b_err;
    _b_err = m_RightOp->get_null_record( records );
    return _b_err;
}

/* ******************** PRIVATE MEMBERS ******************** */
/*!
 *
 */
tgg00_BasisError Join_JoinOperator2::operator_init()
{
    // extract join info and adjust messblock for BD calls
    SAPDBTRACE_METHOD_DEBUG( "Join_JoinOperator2::operator_init", Join_Trace, 3 );

    SAPDB_MemCopyNoCheck( &m_JoinInfo, 
            (SAPDB_Char *)m_acv.a_mblock.mb_strat() + 
            ((*m_acv.a_mblock.mb_st())[m_acv.a_mblock.mb_qual()->mstrat_pos()+2-1].epos() - 1 ), 
            (*m_acv.a_mblock.mb_st())[m_acv.a_mblock.mb_qual()->mstrat_pos()+2-1].elen_var() );
    SAPDBTRACE_IF( Join_Trace, 3, 
        a683trace_joininfo( td_always, m_d_joins, m_JoinInfo ));

    if ( m_IsTopJoinOp )
    {
        m_OutputInfoSize = *(tsp00_Int2*)((SAPDB_Char *) m_acv.a_mblock.mb_strat() + 
            ((*m_acv.a_mblock.mb_st())[ m_acv.a_mblock.mb_qual()->mstrat_pos()+1-1].epos() - 1 ));
        SAPDBTRACE_WRITELN( Join_Trace, 3, "m_OutputInfoSize: " << m_OutputInfoSize );
        if ( m_OutputInfo = (tak68_outcol*) 
                    this->GetAlloc().Allocate( m_OutputInfoSize * sizeof( tak68_outcol ) ))
        {
            SAPDBTRACE_WRITELN( Join_Trace, 3, "allocate m_OutputInfo @ " << (void*)m_OutputInfo );
            SAPDB_MemCopyNoCheck( m_OutputInfo, (SAPDB_Char *) m_acv.a_mblock.mb_strat() + 
                    ((*m_acv.a_mblock.mb_st())[m_acv.a_mblock.mb_qual()->mstrat_pos()+1-1].epos() - 1 ) + 
                    sizeof( tsp00_Int2 ), 
                    m_OutputInfoSize * sizeof( tak68_outcol ) );
#ifdef SAPDB_SLOW
            for ( tsp00_Int2 _i = 0; _i < m_OutputInfoSize; ++_i )
            {
                SAPDBTRACE_WRITELN( Join_Trace, 3, 
                           "oc_srctab: "   << (m_OutputInfo+_i)->oc_srctab 
                        << "\toc_srcpos: " << (m_OutputInfo+_i)->oc_srcpos 
                        << "\toc_dstpos: " << (m_OutputInfo+_i)->oc_dstpos 
                        << "\toc_len: "    << (m_OutputInfo+_i)->oc_len << NewLine );
            }
#endif
        }
        else
            return e_no_more_memory;

        // reserve extra space (failure tolerance)
        if ( this->join_reclen() > this->GetRecordLength() )
            m_JoinRecBufferSize = this->join_reclen() + 20;
        else
            m_JoinRecBufferSize = this->GetRecordLength() + 20;
        if (!( m_JoinRec = (tgg00_Rec*) this->GetAlloc().Allocate( m_JoinRecBufferSize )))
            return e_no_more_memory;
        else
        {
            SAPDBTRACE_WRITELN( Join_Trace, 3, m_JoinRecBufferSize << " bytes for join buffer allocated" );
            SAPDBTRACE_WRITELN( Join_Trace, 3, "allocate m_JoinRec @ " << (void*)m_JoinRec );
        }
        tgg00_BasisError _b_err = e_ok;
        if ( m_acv.a_mblock.mb_qual()->mresqual_pos() > 0 && m_acv.a_mblock.mb_qual()->mresqual_cnt() > 0 )
            _b_err = create_filter( st_jump_output == (*m_acv.a_mblock.mb_st())[m_acv.a_mblock.mb_qual()->mresqual_pos()-1].etype() );
        if  ( e_ok != _b_err ) return _b_err;
        
        if ( ! ( m_Records = (tgg00_Rec**) 
                this->GetAlloc().Allocate( m_RightTableIdx * sizeof(tgg00_Rec*)) ) )
            return e_no_more_memory;

    }
    
    if ( a70glob_join_strats.includes( m_RightAccessStrat ) )
    {
        SAPDB_Int4 _JoinLenPart = 0;
        for ( int _i = 1; _i <= m_JoinInfo.jn_jarr_path; ++_i )
        {
            _JoinLenPart += 
                m_d_joins.jrc_joinarr[ m_JoinInfo.jn_jarr[ _i-1 ] ].jo_recs[ m_d_joins.jrc_joinarr[ m_JoinInfo.jn_jarr[ _i-1 ] ].jo_partno-1 ].jop_inoutlen;
        }
        // get information on changing data types ascii<->ebcdic before joining
        if ( a70glob_join_key_strats.includes( m_RightAccessStrat ) )
        {
            k74check_conversions( m_acv.a_mblock,
                      m_RightRecConvArr, m_RightRecConvArrCnt,
                      m_LastRightDefBytePos, _JoinLenPart);
        }
        else
        {
            k74check_inv_conversions( m_acv.a_mblock, 
                      m_RightRecConvArr, m_RightRecConvArrCnt, 
                      m_LastRightDefBytePos, m_IsDescendingIndex, 
                      _JoinLenPart );
        }
    }
    else
        k74get_last_field_pos( m_acv.a_mblock, 
                  m_LastRightDefBytePos, m_JoinInfo.jn_jarr_path );

    if ( m_JoinInfo.jn_rightoj )
    {
        // determine auxiliary RIGHT OUTER JOIN file name
        g04build_temp_tree_id( m_AuxRightOuterJoinFile, *m_acv.a_mblock.mb_trns() );
        m_AuxRightOuterJoinFile.fileTfnTemp_gg00().becomes( ttfnOuterJoin_egg00 );
        m_AuxRightOuterJoinFile.fileTempCnt_gg00() = m_RightTableIdx;
        m_AuxRightOuterJoinFilePos.tpsPno_gg00     = NIL_PAGE_NO_GG00;
        m_AuxRightOuterJoinKey.len() = 0;
        m_AuxRightOuterKeyLen = get_AuxRightOuterKeyLen();
        SAPDBTRACE_WRITELN( Join_Trace, 3, "m_AuxRightOuterKeyLen : " << m_AuxRightOuterKeyLen );
        
    }
    return m_acv.a_mblock.mb_trns()->trError_gg00;
}

/*!
 *
 */
tgg00_BasisError Join_JoinOperator2::create_filter( SAPDB_Bool is_outputfilter )
{
    SAPDBTRACE_METHOD_DEBUG( "Join_JoinOperator2::create_filter", Join_Trace, 3 );
    
    tgg00_BasisError _b_err = e_ok;
    m_FilterDesc = m_acv.a_mblock.mb_qual()->mstack_desc();
    if ( m_Filter = (tgg00_StackList*) 
                this->GetAlloc().Allocate( m_acv.a_mblock.mb_qual()->mresqual_cnt() * sizeof( tgg00_StackEntry ) ))
    {
        SAPDBTRACE_WRITELN( Join_Trace, 3, "allocate m_Filter @ " << (void*)m_Filter );
        SAPDB_MemCopyNoCheck( m_Filter, (tgg00_StackEntry *)m_acv.a_mblock.mb_st() + ( m_acv.a_mblock.mb_qual()->mresqual_pos() - 1 ), 
                m_acv.a_mblock.mb_qual()->mresqual_cnt() * sizeof( tgg00_StackEntry ) );
    }
    else
        _b_err = e_no_more_memory;

    tgg00_DataPartPtr _datapart; // deallocating as sfp_data_addr()!!!
    if (_datapart = (tgg00_DataPartPtr) this->GetAlloc().Allocate( m_acv.a_mblock.mb_data_len()))
    {
        SAPDBTRACE_WRITELN( Join_Trace, 3, "allocate m_SelFields.sfp_data_addr @ " << (void*)_datapart );
        SAPDB_MemCopyNoCheck( _datapart, m_acv.a_mblock.mb_data(), m_acv.a_mblock.mb_data_len() );
    }
    else
        _b_err = e_no_more_memory;

    m_FilterDesc.mst_addr()         = m_Filter;
    m_FilterDesc.mqual_cnt = m_acv.a_mblock.mb_qual()->mresqual_cnt();
    m_FilterDesc.mqual_pos = 1;
    m_FilterDesc.mst_max() = m_FilterDesc.mqual_cnt;
    m_FilterDesc.mfirst_free() = m_FilterDesc.mqual_cnt + 1;
    m_FilterDesc.mst_optimize_pos() = 0;
    m_FilterDesc.mstrat_pos         =  m_FilterDesc.mstrat_cnt     = 0;
    m_FilterDesc.mresqual_pos()     =  m_FilterDesc.mresqual_cnt() = 0;
    m_FilterDesc.mcol_pos           =  m_FilterDesc.mcol_cnt       = 0;
    m_FilterDesc.mmult_pos          =  m_FilterDesc.mmult_cnt      = 0;
    m_FilterDesc.mview_pos          =  m_FilterDesc.mview_cnt      = 0;
    m_FilterDesc.mupd_pos           =  m_FilterDesc.mupd_cnt       = 0;
    m_FilterDesc.mlink_pos          =  m_FilterDesc.mlink_cnt      = 0;
    m_FilterDesc.mstring_pos        =  m_FilterDesc.mstring_cnt    = 0;
    m_FilterDesc.mtrigger_pos       =  m_FilterDesc.mtrigger_cnt   = 0;
    m_FilterDesc.minvqual_pos       =  m_FilterDesc.minvqual_cnt   = 0;

    SAPDBTRACE_IF( Join_Trace, 3, 
    t01stackdesc( td_always, "copied rqual", m_FilterDesc.mst_addr(), m_FilterDesc ));

    g04init_select_fields( m_SelFields, 
            reinterpret_cast<tsp00_MoveObjPtr>(_datapart), // write into m_SelFields.sfp_data_addr() 
            m_acv.a_mblock.mb_data_len(), 
            m_acv.a_mblock.mb_valuearr(),
            m_acv.a_mblock.mb_validx_max(), 
            m_acv.a_work_st_addr, 
            m_acv.a_work_st_max, 
            m_acv.a_work_buf_addr, 
            m_acv.a_work_buf_size,
            m_acv.a_sqlmode );
    m_SelFields.sfp_bd_mess_type().becomes( m_select );
    m_SelFields.sfp_result_wanted() = true;
    m_SelFields.sfp_m_result_size() = m_JoinRecBufferSize;
    m_SelFields.sfp_first_qual()      = true;
    m_SelFields.sfp_rows_read()       = 0;
    m_SelFields.sfp_rows_qual()       = 0;
    m_SelFields.sfp_acv_addr()        = reinterpret_cast<tsp00_Addr>( &m_acv );
    m_SelFields.sfp_oldrec_addr()     = 0;
    m_SelFields.sfp_primkey_addr()    = 0;
    //SAPDBTRACE_IF( Join_Trace, 3, t01selfields( td_always, "m_SelFields ", m_SelFields ));
    SAPDBTRACE_WRITELN( Join_Trace, 3, "is output filter " << is_outputfilter );
    m_OutputFilter = is_outputfilter;

    /* remove result qualification from message block */
    m_acv.a_mblock.mb_qual()->mfirst_free() = m_acv.a_mblock.mb_qual()->mresqual_pos();
    m_acv.a_mblock.mb_qual()->mresqual_pos() = m_acv.a_mblock.mb_qual()->mresqual_cnt() = 0;

    if ( m_acv.a_mblock.mb_qual()->msubquery() )
    {
        tgg00_Rec* _rec = reinterpret_cast<tgg00_Rec*>( this->GetAlloc().Allocate( sizeof(tgg00_Rec) ) );;
        if ( _rec )
        {
            k720_test_subquery( *m_acv.a_mblock.mb_trns(), 
                *_datapart, m_acv.a_mblock.mb_data_len(),
                m_FilterDesc, *_rec );
            _b_err = m_acv.a_mblock.mb_trns()->trError_gg00;
            this->GetAlloc().Deallocate( _rec );
        }
        else
            _b_err = e_no_more_memory;
    }
    return _b_err;
}

/*!
 *
 */
tgg00_BasisError Join_JoinOperator2::prepare_right_key( tgg00_Rec* *records )
{
    SAPDBTRACE_METHOD_DEBUG( "Join_JoinOperator2::prepare_right_key", Join_Trace, 3 );
    SAPDBERR_ASSERT_STATE( a70glob_join_strats.includes( m_RightAccessStrat ) );
    SAPDBERR_ASSERT_STATE( m_JoinInfo.jn_jarr_path > 0 );
        
    SAPDBTRACE_WRITELN( Join_Trace, 5, "prepare for access: " << m_RightAccessStrat );


    tak_one_join _joinpred;
    SAPDB_Int2 _src_tabno, _src_side; 

    if ( a70glob_join_key_strats.includes( m_RightAccessStrat ) )
    {
        // set primary startkey
        m_Startkeys.reckey.keyLen_gg00() = 0;
        for ( SAPDB_UInt2 _i = 1; _i <= m_JoinInfo.jn_jarr_path; ++_i )
        {
            // copy key parts from different tables
            SAPDBTRACE_IF( Join_Trace, 3, 
            a683_one_join_entry( td_always, m_d_joins, m_JoinInfo.jn_jarr[ _i - 1 ], false ));
            _joinpred = m_d_joins.jrc_joinarr[ m_JoinInfo.jn_jarr[ _i - 1 ] ];
            if ( 0 == (_joinpred.jo_partno-1) /* _dst_side */ ) 
                _src_side = 1; 
            else 
                _src_side = 0;
            SAPDBTRACE_WRITELN( Join_Trace, 5, "_src_side: " << _src_side << " _dst_side: " << _joinpred.jo_partno - 1 );
            
            if ( cak68_join_value == _joinpred.jo_recs[ _src_side ].jop_tableno 
                ||
                 cak68_join_value == _joinpred.jo_recs[ _joinpred.jo_partno-1 ].jop_tableno )
            {
                _src_tabno = m_JoinSequence[m_RightTableIdx-1-1].jos_source;
                SAPDBTRACE_WRITELN( Join_Trace, 5, "get value from " << _src_tabno );
            }
            else
            {
                _src_tabno = _joinpred.jo_recs[ _src_side ].jop_tableno;
                SAPDBTRACE_WRITELN( Join_Trace, 5, "_src_tabno: " << _src_tabno );
            }
            SAPDBERR_ASSERT_STATE( cak68_join_value != _src_tabno );
                
            SAPDB_MemCopyNoCheck( 
                    (SAPDB_Byte*)(&m_Startkeys.reckey.keyVal_gg00()) + 
                    m_Startkeys.reckey.keyLen_gg00(), 
                    ((SAPDB_Byte*)(*(records+(_src_tabno-1)))) + 
                    _joinpred.jo_recs[ _src_side ].jop_outpos - 1, 
                    _joinpred.jo_recs[ _src_side ].jop_inoutlen );
            SAPDBTRACE_WRITELN( Join_Trace, 7, 
                    SAPDBTrace_HexBuf((SAPDB_Byte*)(&m_Startkeys.reckey.keyVal_gg00()) + m_Startkeys.reckey.keyLen_gg00(), _joinpred.jo_recs[ _src_side ].jop_inoutlen ));
            m_Startkeys.reckey.keyLen_gg00() += 
                _joinpred.jo_recs[ _src_side ].jop_inoutlen;
        }
        // conversions not neccessary
        switch ( m_RightAccessStrat )
        {
            // complete key access
            case strat_join_viewkey:
            case strat_join_key_equal:
            case strat_join_all_keys_equal:
                // last key is cutted off
                SAPDB_Char _defbyte; 
//                _defbyte = *((SAPDB_Char*)*(records+(_src_tabno-1)) + 
//                            _joinpred.jo_recs[ _src_side ].jop_outpos - 1 );
                _defbyte = m_Startkeys.reckey.keyVal_gg00()[ (m_Startkeys.reckey.keyLen_gg00()-_joinpred.jo_recs[ _src_side ].jop_inoutlen+1) - 1 ];
                SAPDBTRACE_WRITELN( Join_Trace, 5, "last defbyte: " << SAPDBTrace_Hex(&_defbyte, 1) << ", act key len: " << m_Startkeys.reckey.keyLen_gg00() );
                m_Startkeys.reckey.keyLen_gg00() -= 
                    // last inoutlen
                    _joinpred.jo_recs[ _src_side ].jop_inoutlen -
                    ( s30lnr_defbyte( &m_Startkeys.reckey.keyVal_gg00(), _defbyte, 
                        (m_Startkeys.reckey.keyLen_gg00()-_joinpred.jo_recs[ _src_side ].jop_inoutlen+1) + 1,
                        _joinpred.jo_recs[ _src_side ].jop_inoutlen - 1 ) 
                    + 1 /* defbyte lenght */);
                SAPDBTRACE_WRITELN( Join_Trace, 5, "key len: " << m_Startkeys.reckey.keyLen_gg00() );
                // set primary stopkey
                m_Stopkeys.reckey.keyLen_gg00() = m_Startkeys.reckey.keyLen_gg00();
                SAPDB_MemCopyNoCheck( &m_Stopkeys.reckey.keyVal_gg00(), 
                        &m_Startkeys.reckey.keyVal_gg00(), 
                        m_Stopkeys.reckey.keyLen_gg00() );
                break;
            // key part access
            case strat_join_key_next:
            case strat_join_key_range:
                // set primary startkey, already done
                // set primary stopkey
                m_Stopkeys.reckey.keyLen_gg00() = 
                       sizeof( m_Stopkeys.reckey.keyVal_gg00() );
                SAPDB_MemCopyNoCheck( &m_Stopkeys.reckey.keyVal_gg00(), 
                        &m_Startkeys.reckey.keyVal_gg00(), 
                        m_Startkeys.reckey.keyLen_gg00() );
                // fill stopkey with 0xFF
                SAPDB_MemFillNoCheck( ((SAPDB_Byte*)&m_Stopkeys.reckey.keyVal_gg00()) + 
                        m_Startkeys.reckey.keyLen_gg00(), 0xFF, 
                        sizeof(m_Stopkeys.reckey.keyVal_gg00()) - 
                        m_Startkeys.reckey.keyLen_gg00() );
                break;
            default:
                SAPDBTRACE_WRITELN( Join_Trace, 1, "unknown strategy: " << m_RightAccessStrat << NewLine << __FILE__ << " line " << __LINE__ );
                return e_unknown_strategy;
        }
    }
    else if ( a70glob_join_inv_strats.includes( m_RightAccessStrat ) )
    {
        // set index startkey
        m_Startkeys.listkey.keyLen_gg00() = 0;
        for ( SAPDB_UInt2 _i = 1; _i <= m_JoinInfo.jn_jarr_path; ++_i )
        {
            // copy key parts from different tables
            SAPDBTRACE_IF( Join_Trace, 3, 
            a683_one_join_entry( td_always, m_d_joins, m_JoinInfo.jn_jarr[ _i - 1 ], false ));
            _joinpred = m_d_joins.jrc_joinarr[ m_JoinInfo.jn_jarr[ _i - 1 ] ];
            if ( 0 == (_joinpred.jo_partno-1) /* _dst_side */ ) 
                _src_side = 1; 
            else 
                _src_side = 0;
            SAPDBTRACE_WRITELN( Join_Trace, 5, "_src_side: " << _src_side << " _dst_side: " << _joinpred.jo_partno-1 );

            if ( cak68_join_value == _joinpred.jo_recs[ _src_side ].jop_tableno 
                ||
                 cak68_join_value == _joinpred.jo_recs[ _joinpred.jo_partno-1 ].jop_tableno )
            {
                _src_tabno = m_JoinSequence[m_RightTableIdx-1-1].jos_source;
                SAPDBTRACE_WRITELN( Join_Trace, 5, "get value from " << _src_tabno );
            }
            else
            {
                _src_tabno = _joinpred.jo_recs[ _src_side ].jop_tableno;
                SAPDBTRACE_WRITELN( Join_Trace, 5, "_src_tabno: " << _src_tabno );
            }
            SAPDBERR_ASSERT_STATE( cak68_join_value != _src_tabno );

            SAPDB_MemCopyNoCheck( 
                    (SAPDB_Byte*)&m_Startkeys.listkey.keyVal_gg00() + 
                    m_Startkeys.listkey.keyLen_gg00(), 
                    ((SAPDB_Byte*)*(records+(_src_tabno-1))) + 
                    _joinpred.jo_recs[ _src_side ].jop_outpos - 1, 
                    _joinpred.jo_recs[ _src_side ].jop_inoutlen );
            SAPDBTRACE_WRITELN( Join_Trace, 7, 
                    SAPDBTrace_HexBuf((SAPDB_Byte*)(&m_Startkeys.listkey.keyVal_gg00()) + m_Startkeys.listkey.keyLen_gg00(), _joinpred.jo_recs[ _src_side ].jop_inoutlen ));
            m_Startkeys.listkey.keyLen_gg00() += 
                _joinpred.jo_recs[ _src_side ].jop_inoutlen;
        }
        switch ( m_RightAccessStrat )
        {
            // complete index access
            case strat_join_inv:
            case strat_join_all_inv_equal:
                // last key is cutted off
                SAPDB_Char _defbyte; 
                _defbyte = m_Startkeys.listkey.keyVal_gg00()[ (m_Startkeys.listkey.keyLen_gg00()-_joinpred.jo_recs[ _src_side ].jop_inoutlen+1) - 1 ];
                SAPDBTRACE_WRITELN( Join_Trace, 5, "last defbyte: " << SAPDBTrace_Hex(&_defbyte, 1) << ", act key len: " << m_Startkeys.listkey.keyLen_gg00() );
                m_Startkeys.listkey.keyLen_gg00() -= 
                    // last inoutlen
                    _joinpred.jo_recs[ _src_side ].jop_inoutlen -
                    ( s30lnr_defbyte( &m_Startkeys.listkey.keyVal_gg00(), _defbyte, 
                        (m_Startkeys.listkey.keyLen_gg00()-_joinpred.jo_recs[ _src_side ].jop_inoutlen+1) + 1,
                        _joinpred.jo_recs[ _src_side ].jop_inoutlen - 1 ) 
                    + 1 /* defbyte lenght */);
                SAPDBTRACE_WRITELN( Join_Trace, 5, "build inv len: " << m_Startkeys.listkey.len()
                        << "\tis descending: " << m_IsDescendingIndex 
                        << "\nindex length: " << m_RightAccessInvLen );
                // ascii/ebcdic, descending conversion
                convert_key( m_Startkeys.listkey );
                if ( m_IsDescendingIndex && 
                     m_RightAccessInvLen > m_Startkeys.listkey.len() )
                {
                    // we have to fill invkey for descending indexes with define 
                    // bytes because descending key are blown up to full length
                    if ( (0xFF - csp_unicode_def_byte) == m_Startkeys.listkey.k()[ m_LastRightDefBytePos - 1 - cgg_rec_key_offset ] )
                    {
                        g20unifill( sizeof(m_Startkeys.listkey.k()), 
                                &m_Startkeys.listkey.k(), 
                                m_Startkeys.listkey.len()+1,
                                m_RightAccessInvLen - m_Startkeys.listkey.len(),
                                csp_unicode_blank );
                        for ( int _ix = m_Startkeys.listkey.len() + 1; 
                              _ix <= m_RightAccessInvLen; ++_ix )
                            m_Startkeys.listkey.k()[ _ix - 1 ] = 0xFF - m_Startkeys.listkey.k()[ _ix - 1 ];
                    }
                    else
                    {
                        SAPDB_MemFillNoCheck( (SAPDB_Byte*)&m_Startkeys.listkey.k() + m_Startkeys.listkey.len(), 
                                m_Startkeys.listkey.k()[ m_LastRightDefBytePos - cgg_rec_key_offset - 1 ], 
                                m_RightAccessInvLen - m_Startkeys.listkey.len() );
                    }
                    m_Startkeys.listkey.len() = m_RightAccessInvLen;
                    SAPDBTRACE_WRITELN( Join_Trace, 5, "set listkey.len(): " << m_Startkeys.listkey.len() );
                }
                // set index stopkey
                m_Stopkeys.listkey.len() = m_Startkeys.listkey.len();
                SAPDB_MemCopyNoCheck( &m_Stopkeys.listkey.k(), 
                        &m_Startkeys.listkey.k(), m_Startkeys.listkey.len() );
                break;
            // index part access
            case strat_join_inv_range:
                // set index startkey, already done
                // ascii/ebcdic, descending conversion
                convert_key( m_Startkeys.listkey );
                // set index stopkey
                m_Stopkeys.listkey.len() = sizeof(m_Stopkeys.listkey.k());
                SAPDB_MemCopyNoCheck( &m_Stopkeys.listkey.k(), &m_Startkeys.listkey.k(), 
                        m_Startkeys.listkey.len() );
                SAPDB_MemFillNoCheck( ((SAPDB_Byte*)&m_Stopkeys.listkey.keyVal_gg00()) + 
                        m_Startkeys.listkey.keyLen_gg00(), 0xFF, 
                        sizeof(m_Stopkeys.listkey.keyVal_gg00()) - 
                        m_Startkeys.listkey.keyLen_gg00() );
                break;
            default:
                SAPDBTRACE_WRITELN( Join_Trace, 1, "unknown strategy: " << m_RightAccessStrat << NewLine << __FILE__ << " line " << __LINE__ );
                return e_unknown_strategy;
        }

        // set primary start/stopkey
        m_Startkeys.reckey.len() = 0;
        m_Stopkeys.reckey.len() = 0;
    }

    // too long key could return e_no_next_record
    return e_ok;
}

/*!
 *
 */
SAPDB_Bool Join_JoinOperator2::join( tgg00_Rec* *records, tgg00_BasisError &b_err ) 
{
    SAPDBTRACE_METHOD_DEBUG( "Join_JoinOperator2::join", Join_Trace, 3 );

    SAPDB_UInt2 _joincnt = 1;
    SAPDB_Bool _fullfilled = true;
    tak_one_join _joinpred;

    while ( _joincnt <= m_JoinInfo.jn_jarr_cnt && _fullfilled && e_ok == b_err )
    {
        SAPDBTRACE_IF( Join_Trace, 3, 
        a683_one_join_entry( td_always, m_d_joins, m_JoinInfo.jn_jarr[ _joincnt - 1 ], false ));
        SAPDB_Bool _reverse_operands = false;
        _joinpred = 
            m_d_joins.jrc_joinarr[ m_JoinInfo.jn_jarr[ _joincnt - 1 ] ];
        tgg00_StackOpType _stackop = _joinpred.jo_op;

        SAPDBTRACE_IF( Join_Trace, 3, 
            t01op( td_always, "_stackop    ", _stackop ));

        if  ( 1 == _joinpred.jo_partno )
        {
            switch ( _stackop )
            {
                case op_lt:
                    _stackop.becomes( op_gt );
                    break;
                case op_le:
                    _stackop.becomes( op_ge );
                    break;
                case op_gt:
                    _stackop.becomes( op_lt );
                    break;
                case op_ge:
                    _stackop.becomes( op_le );
                    break;
                default:
                    /* for op_like, op_not_like, op_sounds, op_not_sounds
                     * _reverse_operands has no effect for op_eq, op_ne   */
                    _reverse_operands = true;
            }
            SAPDBTRACE_IF( Join_Trace, 3, 
                t01op( td_always, "_stackop rev", _stackop ));
        }

        SAPDB_UInt2 _dst_side, _src_side, _src_tabno, _dst_tabno;
        if ( 0 == ( _dst_side = _joinpred.jo_partno-1 ) ) 
            _src_side = 1;
        else
            _src_side = 0;
        SAPDBTRACE_WRITELN( Join_Trace, 5, "_src_side: " << _src_side << " _dst_side: " << _dst_side );
        if ( cak68_join_value == _joinpred.jo_recs[ _src_side ].jop_tableno 
            ||
             cak68_join_value == _joinpred.jo_recs[ _joinpred.jo_partno-1 ].jop_tableno )
        {
            _src_tabno = m_JoinSequence[m_RightTableIdx-1-1].jos_source;
            _dst_tabno = m_JoinSequence[m_RightTableIdx-1  ].jos_source;
            SAPDBTRACE_WRITELN( Join_Trace, 5, "get value from " << _src_tabno << " join with " << _dst_tabno );
        }
        else
        {
            _src_tabno = _joinpred.jo_recs[ _src_side ].jop_tableno;
            _dst_tabno = _joinpred.jo_recs[ _dst_side ].jop_tableno;
            SAPDBTRACE_WRITELN( Join_Trace, 5, "_src_tabno: " << _src_tabno << "\t_dst_tabno: " << _dst_tabno );
        }
        SAPDBERR_ASSERT_STATE( cak68_join_value != _src_tabno && cak68_join_value != _dst_tabno );
        
        SAPDBTRACE_WRITELN( Join_Trace, 5, 
                " join tab " << _src_tabno << 
                "(" << _joinpred.jo_recs[ _src_side ].jop_fieldno << ")" << 
                " on pos " << _joinpred.jo_recs[ _src_side ].jop_outpos <<
                " with " << _dst_tabno << 
                "(" << _joinpred.jo_recs[ _dst_side ].jop_fieldno << ")" << 
                " on pos " << _joinpred.jo_recs[ _dst_side ].jop_outpos );
        
        tsp00_LcompResult _lc_result;
        
        // compare columns
        switch ( _stackop )
        {
            case op_like:
            case op_not_like:
                {
                    tgg00_Rec *_like_pattern = 0;
                    SAPDB_Int4 _change_pos, _ix;
                    if ( _reverse_operands )
                    {
                        SAPDBTRACE_WRITELN( Join_Trace, 5, "reverse operands " << "\tleft record changed: " << m_LeftRecChanged );
                        if ( ! m_LeftRecChanged )
                        {
                            m_LeftRecChanged = true;
                            // left record
                            _like_pattern = *(records+_src_tabno-1);
                            _change_pos   = _joinpred.jo_recs[ _src_side ].jop_outpos;
                            _ix           = 0;
                        }
                    }
                    else
                    {
                        // right record
                        _like_pattern = *(records+_dst_tabno-1);
                        _change_pos   = _joinpred.jo_recs[ _dst_side ].jop_outpos;
                        _ix           = 1;
                    }
                    pasbool _ok;
                    if ( _like_pattern )
                    {
                        if ( csp_unicode_def_byte == _like_pattern->recBuf_gg00()[ _change_pos - 1 ] )
                        {
                            SAPDBTRACE_WRITELN( Join_Trace, 5, "uni build pattern" );
                            s49uni_build_pattern(
                                  _like_pattern, _change_pos + 1,
                                  _change_pos + 
                                  _joinpred.jo_recs[ _ix ].jop_inoutlen,
                                  (char *)csp_unicode_blank, false,
                                  sqlm_ansi, _ok );
                        }
                        else
                        {
                            SAPDBTRACE_WRITELN( Join_Trace, 5, "build pattern" );
                            s49build_pattern(
                                  _like_pattern,
                                  ( csp_ascii_blank == _like_pattern->buf()[ _change_pos - 1 ] ),
                                  _change_pos + 1, _change_pos + 
                                  _joinpred.jo_recs[ _ix ].jop_inoutlen,
                                  bsp_c1, false, false,
                                  sqlm_ansi, _ok );
                        }
                        if ( ! _ok )
                        {
                            if ( sqlm_ansi == m_acv.a_mblock.mb_qual()->msqlmode() )
                                m_acv.a_mblock.mb_trns()->trError_gg00 = e_illegal_escape_sequence;
                            else
                                m_acv.a_mblock.mb_trns()->trError_gg00 = e_invalid_pattern;
                        }
                    }
                }
                // fall through
            case op_sounds:
            case op_not_sounds:
                {
                    SAPDB_Int4 _compare_ok = 0;
                    tkb07_buffer_description _workbuf;

                    _workbuf.buffer_addr = m_acv.a_work_buf_addr;
                    _workbuf.buffer_size = m_acv.a_work_buf_size;
                    _workbuf.buffer_len  = 0;
                    if ( _reverse_operands )
                    {
                        k71join_comparison(m_SelFields, _stackop, _workbuf,
                              /* right record + pos */
                              *(records+_dst_tabno-1),
                              _joinpred.jo_recs[ _dst_side ].jop_outpos,
                              _joinpred.jo_recs[ _dst_side ].jop_inoutlen,
                              /* left record + pos */
                              *(records+_src_tabno-1),
                              _joinpred.jo_recs[ _src_side ].jop_outpos,
                              _joinpred.jo_recs[ _src_side ].jop_inoutlen,
                              _compare_ok );
                    }
                    else
                    {
                        k71join_comparison(m_SelFields, _stackop, _workbuf,
                              /* left record + pos */
                              *(records+_src_tabno-1),
                              _joinpred.jo_recs[ _src_side ].jop_outpos,
                              _joinpred.jo_recs[ _src_side ].jop_inoutlen,
                              /* right record + pos */
                              *(records+_dst_tabno-1),
                              _joinpred.jo_recs[ _dst_side ].jop_outpos,
                              _joinpred.jo_recs[ _dst_side ].jop_inoutlen,
                              _compare_ok );
                    }
                    _fullfilled = ( cgg04_is_true == _compare_ok );
                }
                break;
            default:
                s30luc( 
                        /* left record + pos */
                        *(records+_src_tabno-1),
                        _joinpred.jo_recs[ _src_side ].jop_outpos,
                        _joinpred.jo_recs[ _src_side ].jop_inoutlen,
                        /* right record + pos */
                        *(records+_dst_tabno-1),
                        _joinpred.jo_recs[ _dst_side ].jop_outpos,
                        _joinpred.jo_recs[ _dst_side ].jop_inoutlen,
                        _lc_result );
                SAPDBTRACE_IF( Join_Trace, 3, 
                    t01comp_result( td_always, "comp result ", _lc_result ));
                switch( _stackop )
                {
                    case op_eq:
                        if ( l_equal == _lc_result )
                            _fullfilled = true;
                        else
                            _fullfilled = false;
                        break;
                    case op_le:
                        if ( l_equal == _lc_result || l_less == _lc_result )
                            _fullfilled = true;
                        else
                            _fullfilled = false;
                        break;
                    case op_lt:
                        if ( l_less == _lc_result )
                            _fullfilled = true;
                        else
                            _fullfilled = false;
                        break;
                    case op_ge:
                        if ( l_equal == _lc_result || l_greater ==_lc_result )
                            _fullfilled = true;
                        else
                            _fullfilled = false;
                        break;
                    case op_gt:
                        if ( l_greater == _lc_result )
                            _fullfilled = true;
                        else
                            _fullfilled = false;
                        break;
                    case op_ne :
                        if ( l_equal != _lc_result && l_undef != _lc_result ) 
                            _fullfilled = true;
                        else
                            _fullfilled = false;
                        break;
                    default:
                        _fullfilled = false;
                        SAPDBTRACE_WRITELN( Join_Trace, 1, "operator " << _stackop << " not implemented: " << __FILE__ << " line " << __LINE__ );
                        b_err = e_not_implemented;
                }
        }
        ++_joincnt;
    }
    SAPDBTRACE_WRITELN( Join_Trace, 3, "is join: " << _fullfilled );
    return _fullfilled;
}

/*!
 *
 */
void Join_JoinOperator2::convert_key( tgg00_Lkey &key )
{
    SAPDBTRACE_METHOD_DEBUG( "Join_JoinOperator2::convert_key", Join_Trace, 3 );
    SAPDBTRACE_WRITELN( Join_Trace, 3, "m_RightRecConvArrCnt " << m_RightRecConvArrCnt );
    SAPDBTRACE_IF( Join_Trace, 5, t01buf( td_always, key.k(), 1, key.len() <= 100 ? key.len() : 100 ));
    
    // take conversions from m_RightRecConvArr and adjust key 
    for ( SAPDB_UInt2 _i = 0; _i < m_RightRecConvArrCnt; ++_i )
    {
        SAPDBTRACE_WRITELN( Join_Trace, 3, "ct_epos: " << m_RightRecConvArr[ _i ].ct_epos << "\tct_elenvar: " << m_RightRecConvArr[ _i ].ct_elenvar << "\tct_codeno: " << m_RightRecConvArr[ _i ].ct_codeno );
        if ( m_RightRecConvArr[ _i ].ct_codeno != 0 )
            s30map( g02codetables.tables[ m_RightRecConvArr[ _i ].ct_codeno - 1 ],
                    key.k(), 
                    m_RightRecConvArr[ _i ].ct_epos - cgg_rec_key_offset,
                    key.k(),
                    m_RightRecConvArr[ _i ].ct_epos - cgg_rec_key_offset,
                    m_RightRecConvArr[ _i ].ct_elenvar );
        if ( m_RightRecConvArr[ _i ].ct_is_desc )
        {
            SAPDBTRACE_WRITELN( Join_Trace, 5, "convert desc" );
            // build revers index key for descending indexes
            for ( SAPDB_Int4 _jx = m_RightRecConvArr[ _i ].ct_epos - cgg_rec_key_offset; 
                  _jx <= m_RightRecConvArr[ _i ].ct_epos + m_RightRecConvArr[ _i ].ct_elenvar - cgg_rec_key_offset - 1; 
                  ++_jx )
            {
                key.k()[ _jx - 1 ] = char(0xFF - key.k()[ _jx - 1 ]);
            }
        } 
    }
    SAPDBTRACE_IF( Join_Trace, 5, t01buf( td_always, key.keyVal_gg00(), 1, key.keyLen_gg00() <= 100 ? key.keyLen_gg00() : 100 ));
}

/*!
 *
 */
void Join_JoinOperator2::create_output( 
        tgg00_Rec  *target,
        tgg00_Rec* *records )
{
    SAPDBTRACE_METHOD_DEBUG( "Join_JoinOperator2::create_output", Join_Trace, 3 );
    SAPDBERR_ASSERT_STATE( target ); 
    SAPDBTRACE_IF( Join_Trace, 5, SAPDB_MemFillNoCheck( target, ':', m_JoinRecBufferSize ) );

    tgg00_BasisError _b_err = e_ok;

    target->recLen_gg00()    = join_reclen();
    target->recKeyLen_gg00() = join_keylen();
    target->recVarcolOffset_gg00() = 0;
    target->recVarcolCnt_gg00()    = 0;
    
    for ( tsp00_Int2 _i = 0; _i < m_OutputInfoSize; ++_i )
    {
        SAPDBTRACE_WRITELN( Join_Trace, 3, 
                (m_OutputInfo+_i)->oc_srctab << " : "  << 
                (m_OutputInfo+_i)->oc_srcpos << " -> " <<
                (m_OutputInfo+_i)->oc_dstpos << " ( "  <<
                (m_OutputInfo+_i)->oc_len    << " )" );
        SAPDB_MemCopyNoCheck( 
                (SAPDB_Byte*)&(target->recBuf_gg00()) + (m_OutputInfo+_i)->oc_dstpos - 1, 
                (SAPDB_Byte*)&(*(records + (m_OutputInfo+_i)->oc_srctab - 1))->recBuf_gg00() +
                (m_OutputInfo+_i)->oc_srcpos - 1, 
                (m_OutputInfo+_i)->oc_len );
    }
    
#ifdef SAPDB_SLOW
    SAPDB_MemFillNoCheck( (SAPDB_Byte*)&target->recKey_gg00().k() + target->recKeyLen_gg00() - RESCNT_MXGG04, 0x99, RESCNT_MXGG04 );
#endif
}
/*!
 *
 */
SAPDB_Bool Join_JoinOperator2::filter( tgg00_Rec *recptr, tgg00_BasisError &b_err ) 
{
    // filter and transform (i.e. functions) record
    SAPDBTRACE_METHOD_DEBUG( "Join_JoinOperator2::filter", Join_Trace, 3 );
    tgg00_StackEntry *_err_st_entry;
    pasbool _unqualified = false;

    m_SelFields.sfp_m_result_addr() = reinterpret_cast<tsp00_MoveObjPtr>(recptr);
    m_SelFields.sfp_rec_addr()      = reinterpret_cast<tsp00_BufAddr>(recptr);
    m_SelFields.sfp_rec_len()       = recptr->recLen_gg00();
    m_SelFields.sfp_rec_key_len()   = recptr->recKeyLen_gg00();
    m_SelFields.sfp_result_length() = 0; 
    m_SelFields.sfp_m_result_len()  = 0;
    m_SelFields.sfp_m_result_cnt()  = 1;
    
    SAPDBTRACE_WRITELN( Join_Trace, 7, "before filter: " << m_OutputFilter );
    SAPDBTRACE_IF( Join_Trace, 7, 
    t01buf( td_always, recptr, 1, recptr->recLen_gg00() <= 300 ? recptr->recLen_gg00() : 300 ));
    if ( m_OutputFilter )
    {
        tsp00_Buf _resbuf;
        SAPDBTRACE_IF( Join_Trace, 5, SAPDB_MemFillNoCheck( &_resbuf, ':', sizeof(_resbuf) ));
        m_SelFields.sfp_m_result_size() = sizeof(_resbuf);
        m_SelFields.sfp_m_result_addr() = reinterpret_cast<tsp00_MoveObjPtr>(&_resbuf);
        SAPDB_Int4 _old_result_len = m_SelFields.sfp_result_length();
        m_acv.a_mblock.mb_trns()->trError_gg00 = e_ok;
        
        k71qual_handling( *m_acv.a_mblock.mb_trns(), m_SelFields, 
                m_FilterDesc.mview_cnt > 0, false, m_FilterDesc,
                _err_st_entry, _unqualified );
        // write result length
        recptr->recLen_gg00() = GetRecordLength();
        recptr->recKeyLen_gg00() = GetKeyLength();

        if ( _unqualified )
            m_SelFields.sfp_result_length() = _old_result_len;
        else
        {
            if ( -1 == (*m_Filter)[ m_FilterDesc.mqual_pos - 1 ].ecol_pos() )
            {
                SAPDBTRACE_WRITELN( Join_Trace, 7, "transfer constant expression" );
                /* we don't have complete output qualification of record */
                /* ak684transfer_const_expressions() */
                /* ak684out_filter_stack()           */
                for ( SAPDB_Int4 _ix = m_FilterDesc.mqual_pos + 1; 
                        _ix <= (*m_Filter)[ m_FilterDesc.mqual_pos - 1 ].epos() + 
                        m_FilterDesc.mqual_pos - 2; ++_ix )
                {
                    if ( st_output == (*m_Filter)[ _ix - 1 ].etype() )
                    {
                        SAPDBTRACE_WRITELN( Join_Trace, 5, "pos: " << (*m_Filter)[ _ix - 1 ].epos() << " len: " << (*m_Filter)[ _ix - 1 ].elen_var());
                        SAPDB_MemCopyNoCheck( 
                                (SAPDB_Byte*)recptr + (*m_Filter)[ _ix - 1 ].epos() - 1, 
                                (SAPDB_Byte*)&_resbuf + (*m_Filter)[ _ix - 1 ].epos() - 1,
                                (*m_Filter)[ _ix - 1 ].elen_var() 
                                );
                    }
                }
            }
            else
            {
                SAPDB_MemCopyNoCheck( (SAPDB_Byte*)recptr + cgg_rec_key_offset,
                        (SAPDB_Byte*)&_resbuf + cgg_rec_key_offset,
                        recptr->recLen_gg00() );
            }
            // write defined result counter
            SAPDBTRACE_IF( Join_Trace, 3, SAPDB_MemFillNoCheck( (SAPDB_Byte*)&recptr->recKey_gg00().k() + recptr->recKeyLen_gg00() - RESCNT_MXGG04, 0x99, RESCNT_MXGG04 ) );
        }
    }
    else
    {
        m_SelFields.sfp_m_result_size() = m_JoinRecBufferSize;
        m_SelFields.sfp_m_result_addr() = reinterpret_cast<tsp00_MoveObjPtr>(recptr);

        k71qual_handling( *m_acv.a_mblock.mb_trns(), m_SelFields, 
                m_FilterDesc.mview_cnt > 0, false, m_FilterDesc,
                _err_st_entry, _unqualified );
        // write result length
        recptr->recLen_gg00() = GetRecordLength();
        recptr->recKeyLen_gg00() = GetKeyLength();
    }
    SAPDBTRACE_WRITELN( Join_Trace, 3, "k71qual_handling: " << SAPDBTrace::BasisError(m_acv.a_mblock.mb_trns()->trError_gg00) );
    
#ifdef SAPDB_SLOW
    if ( e_ok == m_acv.a_mblock.mb_trns()->trError_gg00 )
    {
        SAPDBTRACE_WRITELN( Join_Trace, 7, "after filter: " );
        SAPDBTRACE_IF( Join_Trace, 7, 
        t01buf( td_always, recptr, 1, recptr->recLen_gg00() <= 300 ? recptr->recLen_gg00() : 300 ));
    }
#endif
    
    switch ( m_acv.a_mblock.mb_trns()->trError_gg00 ) 
    {
        case e_ok:
            return true; break;
        case e_qual_violation:
            m_acv.a_mblock.mb_trns()->trError_gg00 = e_ok; 
            return false; break;
        default:
            b_err = m_acv.a_mblock.mb_trns()->trError_gg00;
            return false;
    }
}

/*!
*
*/
void Join_JoinOperator2::init() 
{
    if ( m_JoinInfo.jn_rightoj )
        // create auxiliary RIGHT OUTER JOIN file
        b01tcreate_file( *m_acv.a_mblock.mb_trns(), m_AuxRightOuterJoinFile );
    m_TableForRightOuterJoinScanStat       = never_scanned;
    m_AuxRightOuterJoinFilePos.tpsPno_gg00 = NIL_PAGE_NO_GG00;
    m_AppendRightOuterJoinRecords = false;
    m_AuxRightOuterJoinKey.len()  = 0;
}

/*!
 *
 */
SAPDB_UInt4 Join_JoinOperator2::get_AuxRightOuterKeyLen()
{
    SAPDBTRACE_METHOD_DEBUG( "Join_JoinOperator2::get_AuxRightOuterKeyLen", Join_Trace, 3 );
    SAPDBERR_ASSERT_STATE( st_jump_output == (*m_acv.a_mblock.mb_st())[m_acv.a_mblock.mb_qual()->mqual_pos()-1].etype() );

    SAPDB_Int4 _stop = m_acv.a_mblock.mb_qual()->mqual_pos();
    SAPDB_Int4 _i = (*m_acv.a_mblock.mb_st())[m_acv.a_mblock.mb_qual()->mqual_pos()-1].epos() + 
        m_acv.a_mblock.mb_qual()->mqual_pos() - 1;
    
    // loop over output columns, look for output genertated by ak684key_columns_stack()
    while ( _i > _stop )
    {
        --_i;
        if ( ( st_varkey == (*m_acv.a_mblock.mb_st())[_i-1].etype() || 
               st_fixkey == (*m_acv.a_mblock.mb_st())[_i-1].etype()) &&
             st_output == (*m_acv.a_mblock.mb_st())[(_i-1)+1].etype() &&
             op_o_output_outer_join == (*m_acv.a_mblock.mb_st())[(_i-1)+1].eop_out() )
        {
            SAPDBTRACE_WRITELN( Join_Trace, 5, "VAR KEY at pos " << _i << "\tkey len " << (*m_acv.a_mblock.mb_st())[_i-1].ecol_pos() );
            return (*m_acv.a_mblock.mb_st())[_i-1].ecol_pos();
        }
    }
    a07_b_put_error( m_acv, e_key_does_not_exist, 1 );
    SAPDBERR_ASSERT_STATE( false );
    return 0;
}

IOperator* Join_JoinOperator2::AccessOperatorFactory(
    tak_all_command_glob&          acv,
    const SAPDB_UInt4&             buffersize,
    const SAPDB_UInt4&             expectedRecordCount,
    const SAPDB_UInt2              tabno)
{
    SAPDBTRACE_METHOD_DEBUG( "Join_JoinOperator2::AccessOperatorFactory", Join_Trace, 3 );

    Join_AccessDesc _accessDesc( acv );
    Join_IAccessOperator*     _accessOp     = 0;
    Join_HashAccessOperator*  _hashAccessOp = 0;
    SAPDB_Byte*               _recordBuffer = 0;
    SAPDB_UInt                _recBufSize   = buffersize;

    SAPDBTRACE_WRITELN( Join_Trace, 3, "access strategy : " << _accessDesc.Strategy()->str_strategy );

    tgg00_BasisError _b_err = e_ok;

    if ( ! ( _accessDesc.IsFetchAccess() || _accessDesc.IsJoinAccess() ||
             strat_key_equal == _accessDesc.Strategy()->str_strategy ) )
        _b_err = e_unknown_strategy;

    if ( strat_inv_range_merge_fetch == _accessDesc.Strategy()->str_strategy )
        _b_err = e_not_implemented;

    if ( e_ok == _b_err ) 
    {
        if ( _accessDesc.UseHashJoin() && a01use_join_hashtable ) 
        {
            SAPDB_UInt4 _expectedTableSize =
                static_cast<SAPDB_UInt4>(
                    expectedRecordCount * _accessDesc.GetRecordLength() * 1.1 );
            if ( Join_Trace.TracesLevel( 1 ) )
            {
                SAPDBTrace_Stream traceStream( &Join_Trace.GetTracePrinter() );
                traceStream
                    << "Join_JoinOperator - hash access:" << NewLine
                    << "  Expected record count: " << expectedRecordCount << NewLine
                    << "  Record length        : "
                    << _accessDesc.GetRecordLength() << NewLine
                    << "  Expected table size  : " << _expectedTableSize << NewLine;
                }
            if ( _expectedTableSize < BUF8K_MXSP00 ) 
            {
                SAPDBTRACE_WRITELN(
                    Join_Trace, 1,
                    "Corrected expected table size from "
                    << _expectedTableSize << " to "
                    << BUF8K_MXSP00 << " bytes" );
                _expectedTableSize = BUF8K_MXSP00;
            }

            SAPDB_Int2 _joinTransitionKeyLength = 0;
            for ( int _i = 1; _i <= m_JoinInfo.jn_jarr_path; ++_i )
            {
                _joinTransitionKeyLength += 
                    m_d_joins.jrc_joinarr[ m_JoinInfo.jn_jarr[ _i-1 ] ].jo_recs[ m_d_joins.jrc_joinarr[ m_JoinInfo.jn_jarr[ _i-1 ] ].jo_partno-1 ].jop_inoutlen;
            }

            SAPDB_Int2 _lastKeyColStartPos = m_LastRightDefBytePos
                - cgg_rec_key_offset - 1;
            Join_HashAccessOperator::UniqueKeys _uniqueKeys =
                (_accessDesc.Strategy()->str_strategy == strat_join_key_equal)
                || (_accessDesc.Strategy()->str_strategy == strat_join_all_keys_equal)
                || (_accessDesc.Strategy()->str_strategy == strat_join_inv)
                || (_accessDesc.Strategy()->str_strategy == strat_join_all_inv_equal)
                ? Join_HashAccessOperator::NoDuplicateKeys
                : Join_HashAccessOperator::AllowDuplicateKeys;

            Join_HashAccessOperator::KeySizes _keySizes = 
                (_accessDesc.Strategy()->str_strategy == strat_join_key_next)
                || (_accessDesc.Strategy()->str_strategy == strat_join_key_range)
                ? Join_HashAccessOperator::FixSizedKeys
                : Join_HashAccessOperator::VarSizedKeys;

            _hashAccessOp =
                new (this->GetAlloc()) Join_HashAccessOperator(
                    acv,
                    tabno,
                    _uniqueKeys,
                    _keySizes,
                    _expectedTableSize,
                    _joinTransitionKeyLength,
                    _lastKeyColStartPos);
            if ( _hashAccessOp ) 
            {
                // make access operator use hash record buffer and
                // size if it could be allocated
                _recordBuffer = _hashAccessOp->GetRecordBuffer();
                if ( 0 != _recordBuffer ) {
                    _recBufSize = _expectedTableSize;
                }
            }
        }

        if ( _accessDesc.IsIndexAccess() ) 
        {
            Join_IndexAccessOperator::AccessKind _accessKind =
                _accessDesc.UseParallelIndexAccess()
                ? Join_IndexAccessOperator::ParallelAccess
                : Join_IndexAccessOperator::SerialAccess;

            _accessOp = new (this->GetAlloc())
                Join_IndexAccessOperator(
                    acv,
                    tabno,
                    !m_JoinInfo.jn_rightoj,
                    _accessKind,
                    _recordBuffer,
                    _recBufSize );
        } else {
            _accessOp = new (this->GetAlloc())
                Join_TableAccessOperator(
                    acv,
                    tabno,
                    !m_JoinInfo.jn_rightoj,
                    _recordBuffer,
                    _recBufSize );
        }

        if ( _accessOp 
             && (( strat_join_key_equal
                   == _accessDesc.Strategy()->str_strategy)
                 ||
                 ( strat_join_all_keys_equal
                   == _accessDesc.Strategy()->str_strategy)
                 ||
                 ( strat_join_viewkey
                   == _accessDesc.Strategy()->str_strategy)) ) {
            _accessOp->SetMaximumRowsReadAtATime( 1 );
        }

        if ( _hashAccessOp && _accessOp ) 
        {
            _hashAccessOp->SetAccessOperator( _accessOp );
            _accessOp = _hashAccessOp;
        }

        if ( ! _accessOp ) 
        {
            destroy( _hashAccessOp,  this->GetAlloc() );
            m_acv.a_mblock.mb_trns()->trError_gg00 = e_no_more_memory;
            if ( Join_Trace.TracesLevel( 1 ) )
            {
                SAPDBTrace_Stream traceStream( &Join_Trace.GetTracePrinter() );
                traceStream
                    << "Join_JoinOperator2:: access operator creation failed: no more memory";
            }
        }
    }

    if ( _b_err != e_ok )
        a07_b_put_error( acv, _b_err, 1 );

    return _accessOp;
}

